from kivy.app import App
import traceback


class SettingsInitializer(object):
    """Perform settings initialization on a separate thread."""
    type = 'CEBulk'
    kwargs = None
    basis_func_type = 'polynomial'
    skew_threshold = 40
    app = None
    status = None

    def initialize(self):
        from clease import CEBulk, CECrystal
        try:
            if self.type == 'CEBulk':
                self.app.root.settings = CEBulk(**self.kwargs)
            elif self.type == 'CECrystal':
                self.app.root.settings = CECrystal(**self.kwargs)
            self.status.text = 'Finished initializing'
            self.app.root.settings.basis_func_type = self.basis_func_type
            self.app.root.settings.skew_threshold = self.skew_threshold
        except AssertionError as exc:
            traceback.print_exc()
            msg = "AssertError during initialization " + str(exc)
            self.status.text = msg

        except Exception as exc:
            traceback.print_exc()
            self.status.text = str(exc)



from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty
from kivy.uix.textinput import TextInput
from kivy.uix.spinner import Spinner
from kivy.uix.checkbox import CheckBox
from kivy.uix.popup import Popup
from clease.gui.help_message_popup import HelpMessagePopup
from clease.gui import backup_folder


class FitAlgEditor(FloatLayout):
    close = ObjectProperty(None)
    backup_file = 'fit_alg_editor.txt'
    _pop_up = None

    def __init__(self, **kwargs):
        FloatLayout.__init__(self, **kwargs)
        if not backup_folder.exists():
            backup_folder.mkdir()
        self.load_values()

    @property
    def backup_filename(self):
        return str(backup_folder / self.backup_file)

    def backup(self):
        values = []
        for child in self.ids.mainLayout.children:
            if isinstance(child, (TextInput, Spinner)):
                txt = child.text
                values.append(txt)
            elif isinstance(child, CheckBox):
                values.append(child.active)

        with open(self.backup_filename, 'w') as out:
            for v in values:
                out.write(f"{v}\n")

    def load_values(self):
        try:
            with open(self.backup_filename, 'r') as infile:
                lines = infile.readlines()
            lines = [line.strip() for line in lines]

            counter = 0
            for child in self.ids.mainLayout.children:
                if isinstance(child, (TextInput, Spinner)):
                    child.text = lines[counter]
                    counter += 1
                elif isinstance(child, CheckBox):
                    child.active = lines[counter] == 'True'
                    counter += 1
        except Exception:
            pass

    def dismiss_popup(self):
        self._pop_up.dismiss()
        self._pop_up = None

    def show_help_message(self, msg, title):
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title=title,
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()


class LassoEditor(FitAlgEditor):
    backup_file = 'lasso_editor.txt'


class L2Editor(FitAlgEditor):
    backup_file = 'l2_editor.txt'


class BCSEditor(FitAlgEditor):
    backup_file = 'bcs_editor.txt'


class GAEditor(FitAlgEditor):
    backup_file = 'ga_editor.txt'

    def show_elitism_help(self):
        msg = 'Number of inidividuals that are transferred\n'
        msg += 'to the next generation, without any pairing.\n'
        msg += 'If this is 1, only the best inidividual is\n'
        msg += 'transferred. If it is 2, the two most fit\n'
        msg += 'inidviduals is transferred'
        self.show_help_message(msg, 'Elitism help')

    def show_individual_help(self):
        msg = 'An individual in the CE context is a binary\n'
        msg += 'representation of clusters. If the clusters\n'
        msg += 'are lined up in an array, the individual can be\n'
        msg += '[0, 0, 0, 1, 1, 0, 1, 0, 1, 1] which means that\n'
        msg += 'cluster 4, 5, 7, 8 and 9 is included in the fit\n'
        msg += 'GA tries to find the individual that optimises\n'
        msg += 'the chosen cost function'
        self.show_help_message(msg, 'Individual help')

    def show_max_active_help(self):
        msg = 'If the pool of clusters is large, it might be wise\n'
        msg += 'to limit the maximum of active clusters in the initial\n'
        msg += 'population. This number specifies the maximum allowed\n'
        msg += 'number of 1 in any individual in the initial pool'
        self.show_help_message(msg, 'Max. active pool')

    def show_cost_func_help(self):
        msg = 'Mathematical symbols:\n'
        msg += 'k - number of selected features\n'
        msg += 'n - number of datapoints\n\n'
        msg += 'Cost functions:\n'
        msg += 'AIC   - Afaike Information Criterion (penalty 2*k)\n'
        msg += 'AICC  - Modified AIC (AICC = AIC + (2k^2 + 2k)/(n-k-1)\n'
        msg += 'BIC   - Bayes Information Criterion (penalty k*ln(n))\n'
        self.show_help_message(msg, 'Cost function help')



import logging
from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty, StringProperty
from ase.db import connect
from ase.formula import Formula
from copy import deepcopy
from typing import Tuple, List, Dict
import re
from ase.data import chemical_symbols

logger = logging.getLogger(__name__)

SelectCond = Tuple[List[dict], Dict[str, dict]]

# Kivy can show a black screen if there is too much information in a label
# Limit the view to 100 rows
MAX_ROWS = 100


class DbBrowser(FloatLayout):
    close = ObjectProperty(None)
    text = StringProperty('')

    def __init__(self, **kwargs) -> None:
        self.db_name = kwargs.pop('db_name')
        self.spacing = {
            'id': 2,
            'formula': 7,
            'calc': 10,
            'energy': 6,
            'name': 4,
            'gen': 3,
            'struct_type': 11,
            'size': 4
        }
        self._format_constraints()

        FloatLayout.__init__(self, **kwargs)
        txt = f"| {'id':{self.spacing['id']}} |"
        txt += f" {'formula':{self.spacing['formula']}} |"
        txt += f" {'calculator':{self.spacing['calc']}} |"
        txt += f" {'energy':{self.spacing['energy']}} |"
        txt += f" {'name':{self.spacing['name']}} |"
        txt += f" {'size':{self.spacing['size']}} |"
        txt += f" {'gen':{self.spacing['gen']}} |"
        txt += f" {'struct_type':{self.spacing['struct_type']}} |"
        txt += f" {'converged':9} |"
        self.header = txt
        self.set_rows([], {})

    def _format_row(self, row: dict) -> str:
        rowId = row.get('id', -1)
        formula = row.get('formula', 'unknown')
        calc = row.get('calc', 'none') or 'none'
        name = row.get('name', 'NN') or 'NN'
        gen = int(row.get('gen', 0))
        energy = row.get('energy', '') or 0.0
        str_type = row.get('struct_type', '')
        cnv = int(row.get('converged', 0))
        size = row.get('size', '')
        if isinstance(energy, float):
            energy = f"{energy:.3f}"

        if isinstance(cnv, bool):
            cnv = int(cnv)
        txt = f"| {str(rowId):{self.spacing['id']}} |"
        txt += f" {formula:{self.spacing['formula']}} |"
        txt += f" {calc:{self.spacing['calc']}} |"
        txt += f" {energy:{self.spacing['energy']}} |"
        txt += f" {name:{self.spacing['name']}} |"
        txt += f" {size:{self.spacing['size']}} |"
        txt += f" {str(gen):{self.spacing['gen']}} |"
        txt += f" {str_type:{self.spacing['struct_type']}} |"
        txt += f" {str(cnv):9} |"
        return txt + '\n'

    def _update_spacing(self, row: dict) -> None:
        for k, v in row.items():
            default_length = self.spacing.get(k, 0)

            if isinstance(v, float):
                v = f"{v:.3f}"
            if len(str(v)) > default_length:
                self.spacing[k] = len(str(v))

    def _get_formulas(self, cur) -> Dict[int, str]:
        sql = 'SELECT Z, n, id FROM species'
        cur.execute(sql)
        count_dict = {}
        for item in cur.fetchall():
            rowId = item[2]
            current_count = count_dict.get(rowId, {})
            symb = chemical_symbols[item[0]]
            n = item[1]
            current_count[symb] = n
            count_dict[rowId] = current_count

        formula_dict = {}
        for rowId, count in count_dict.items():
            formula = Formula(_count=count)
            formula_dict[rowId] = formula.format('hill')
        return formula_dict

    def _format_constraints(self) -> None:
        with connect(self.db_name) as db:
            cur = db.connection.cursor()
            formulas = self._get_formulas(cur)
            sql = 'SELECT id, calculator, energy FROM systems'
            cur.execute(sql)
            for item in cur.fetchall():
                row = {
                    'id': item[0],
                    'calc': item[1],
                    'energy': item[2],
                    'formula': formulas[item[0]]
                }
                self._update_spacing(row)

            condition = 'key="name" OR key="size" OR key="struct_type"'
            sql = f'SELECT key, value FROM text_key_values WHERE {condition}'
            cur.execute(sql)
            for item in cur.fetchall():
                self._update_spacing({item[0]: item[1]})

            sql = f'SELECT key, value FROM number_key_values WHERE key="gen"'
            cur.execute(sql)
            for item in cur.fetchall():
                self._update_spacing({item[0]: item[1]})

    def _get_select_conditions(self, txt: str) -> SelectCond:
        system_constraints = []
        kvp_constraints = {}
        prog = re.compile(r"(\w+)[=<>!]*(\w+)")
        known_op = ['=', '>', '<', '>=', '<=', '!=']

        system_cols = [
            'id', 'calculator', 'energy', 'fmax', 'smax', 'natoms', 'user', 'volume', 'mass',
            'charge'
        ]

        # Construct selection conditions
        for cond in txt.split(','):
            cond = cond.strip()
            match = prog.match(cond)
            if match is None:
                continue
            key = match.group(1)
            if key == 'formula':
                # Formula is treated separately
                continue
            value = match.group(2)
            op = cond.replace(key, '')
            op = op.replace(value, '')

            # Check that the reminding part is actually a known operator
            # Should protect against undesired injections
            if op not in known_op:
                logger.warning(f"Unkown operator {cond}")
                continue

            if key in system_cols:
                system_constraints.append({'sql': key + op + '?', 'value': value})

            else:
                default = {'sql': [], 'values': []}
                current_constraint = kvp_constraints.get(key, default)
                current_constraint['sql'].append(f'value{op}?')
                current_constraint['values'].append(value)
                kvp_constraints[key] = current_constraint
        return system_constraints, kvp_constraints

    def on_select(self, txt: str) -> None:
        try:
            system_conditions, kvp_conditions = self._get_select_conditions(txt)
            formula = 'all'
            if 'formula' in txt:
                prog = re.compile(r"formula=(\w+)")
                m = prog.match(txt)
                if m is None:
                    logger.warning("Unsupported operation for formula")
                else:
                    formula = m.group(1)
            self.set_rows(system_conditions, kvp_conditions, formula=formula)
        except Exception as exc:
            logger.exception(exc)

    def _count_kvps(self, cur) -> Dict[int, int]:
        """
        Counts the number of key value pairs for each ID
        """
        tables = ['number_key_values', 'text_key_values']
        counter = {}
        for tab in tables:
            sql = f'select id from {tab}'
            cur.execute(sql)

            for rowId in cur.fetchall():
                counter[rowId] = counter.get(rowId, 0) + 1
        return counter

    def set_rows(self, system_conditions, kvp_conditions, formula='all'):
        all_str = deepcopy(self.header) + '\n'

        rows = {}
        with connect(self.db_name) as db:
            cur = db.connection.cursor()
            formulas = self._get_formulas(cur)

            # Go through the DB and find which IDs are valid according
            # to the constraints
            validIds = set()
            sql = 'SELECT id FROM systems'
            if system_conditions:
                sqls = [x['sql'] for x in system_conditions]
                values = [x['value'] for x in system_conditions]
                sql += ' WHERE ' + ' AND '.join(sqls)
                cur.execute(sql, values)
            else:
                cur.execute(sql)

            for rowId in cur.fetchall():
                validIds.add(rowId[0])

            # Limit valid ids based on kvp_conditions
            for k, v in kvp_conditions.items():
                satisfyConstraint = set()
                tables = ['number_key_values', 'text_key_values']
                for tab in tables:
                    sql = f'select id from {tab} WHERE '
                    sql += 'key=? AND ' + ' AND '.join(v['sql'])
                    cur.execute(sql, [k] + v['values'])
                    for rowId in cur.fetchall():
                        satisfyConstraint.add(rowId[0])
                validIds = validIds.intersection(satisfyConstraint)

            # Apply formula constraint
            if formula != 'all':
                formulaIds = set(k for k, v in formulas.items() if v == formula)
                validIds = validIds.intersection(formulaIds)

            # Extract information from systems that is consistent
            # with validIds
            sql = 'SELECT id, calculator, energy FROM systems'
            cur.execute(sql)
            for item in cur.fetchall():
                rowId = item[0]
                if rowId in validIds:
                    row = {
                        'id': item[0],
                        'formula': formulas[rowId],
                        'calc': item[1],
                        'energy': item[2]
                    }
                    rows[rowId] = row

            # Extract information from number_key_values that is consistent
            # with valid ids
            sql = 'select key, value, id from number_key_values WHERE '
            sql += 'key="gen" OR key="converged"'
            cur.execute(sql)
            for key, value, rowId in cur.fetchall():
                if rowId in validIds:
                    rows[rowId][key] = value

            # Extract information from text_key_values that is consistent
            # with valid ids
            sql = 'select key, value, id from text_key_values WHERE '
            sql += 'key="name" OR key="size" OR key="struct_type"'
            cur.execute(sql)
            for key, value, rowId in cur.fetchall():
                if rowId in validIds:
                    rows[rowId][key] = value

        row_count = 0
        for row in rows.values():
            str_rep = self._format_row(row)
            all_str += str_rep
            row_count += 1
            if row_count >= MAX_ROWS:
                break

        self.text = all_str



import numpy as np

__all__ = ('SVD',)


class SVD:
    """
    Class that stores the singular value decomposition of a matrix X.
    This class can be used if the singular value decomposition of X can
    be re-used many times
    """

    def __init__(self):
        self._U = None  # Right unitary matrix
        self._Vh = None  # Left unititary matrix
        self._S = None  # Singular values

    @property
    def U(self) -> np.ndarray:
        return self._U

    @property
    def Vh(self) -> np.ndarray:
        return self._Vh

    @property
    def S(self) -> np.ndarray:
        return self._S

    def calculate(self, X: np.ndarray) -> None:
        """
        Calculate the SVD of the matrix X

        :param X: Matrix that should be decomposed
        """
        self._U, self._S, self._Vh = np.linalg.svd(X, full_matrices=False)

    def clear(self) -> None:
        """
        Set all stored matrices to None
        """
        self._U = None
        self._Vh = None
        self._S = None

    def has_matrices(self) -> bool:
        """
        Return True if matrices are stored in the attributes
        """
        return (self._U is not None and self._Vh is not None and self._S is not None)



# pylint: disable=undefined-variable
from clease.montecarlo.observers.mc_observer import MCObserver
from clease.montecarlo.observers.sgc_observer import SGCObserver
from clease.montecarlo.observers.corr_func_observer import CorrelationFunctionObserver
from clease.montecarlo.observers.snapshot_observer import Snapshot
from clease.montecarlo.observers.energy_evolution_obs import EnergyEvolution
from clease.montecarlo.observers.site_order_parameter import SiteOrderParameter
from clease.montecarlo.observers.lowest_energy_observer import LowestEnergyStructure
from clease.montecarlo.observers.diffraction_observer import DiffractionObserver
from clease.montecarlo.observers.energy_plot_updater import EnergyPlotUpdater
from clease.montecarlo.observers.concentration_observer import ConcentrationObserver
from clease.montecarlo.observers.multi_state_sgc_observer import (MultiStateSGCConcObserver,
                                                                  SGCState)
from .entropy_prod_rate import *
from .acceptance_rate import *

ADDITIONAL = ("MCObserver", "SGCObserver", "CorrelationFunctionObserver", "Snapshot",
              "EnergyEvolution", "LowestEnergyStructure", "DiffractionObserver",
              "EnergyPlotUpdater", "ConcentrationObserver", "SiteOrderParameter",
              "MultiStateSGCConcObserver", "SGCState")

__all__ = (ADDITIONAL + entropy_prod_rate.__all__ + acceptance_rate.__all__)



from kivy.uix.screenmanager import Screen
from kivy.utils import get_color_from_hex
from kivy.uix.popup import Popup
from kivy.app import App

from clease.gui.constants import FOREGROUND_TEXT_COLOR, INACTIVE_TEXT_COLOR
from clease.gui.load_save_dialog import LoadDialog
import os
import json
from threading import Thread
import traceback


class BaseGenerator(object):
    atoms = None
    generator = None
    status = None
    page = None

    def generate(self):
        if self.status is None:
            msg = 'Generate function needs to be implemented.'
            App.get_running_app().root.ids.status.text = msg


class InitialPoolGenerator(BaseGenerator):

    def generate(self):
        try:
            self.generator.generate_initial_pool(atoms=self.atoms)
            msg = 'Initial pool of structures generated.'
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)
        self.page.structure_generation_in_progress = False


class RandomStructureGenerator(BaseGenerator):

    def generate(self):
        try:
            self.generator.generate_random_structures(atoms=self.atoms)
            msg = 'Random structures generated.'
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
        self.page.structure_generation_in_progress = False


class ProbeStructureGenerator(BaseGenerator):
    Tmin = None
    Tmax = None
    num_temp = None
    num_steps = None

    def generate(self):
        try:
            self.generator.generate_probe_structure(atoms=self.atoms,
                                                    init_temp=self.Tmax,
                                                    final_temp=self.Tmin,
                                                    num_temp=self.num_temp,
                                                    num_steps_per_temp=self.num_steps)
            msg = 'Probe strcutres generated.'
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
        self.page.structure_generation_in_progress = False


class GSStructGenerator(object):
    Tmax = None
    Tmin = None
    num_temps = None
    num_steps = None
    eci = None
    randomize = None

    def generate(self):
        try:
            self.generator.generate_gs_structure(atoms=self.atoms,
                                                 init_temp=self.Tmax,
                                                 final_temp=self.Tmin,
                                                 num_temp=self.num_temps,
                                                 num_steps_per_temp=self.num_steps,
                                                 eci=self.eci,
                                                 random_composition=self.randomize)
            msg = 'Ground-state structures generated.'
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
        self.page.structure_generation_in_progress = False


class InsertStructureCB(object):
    """
    Callable class used to update the status field when inserting
    structures

    :param status: widget
        Status field of the page
    """

    def __init__(self, status):
        self.status = status

    def __call__(self, num, tot):
        self.status.text = f"Inserted {num} of {tot} structures"


class ExaustiveGSStructGenerator(object):
    Tmax = None
    Tmin = None
    num_temps = None
    num_steps = None
    eci = None
    num_templates = None
    num_prim_cells = None

    def generate(self):
        try:
            self.generator.generate_gs_structure_multiple_templates(
                num_templates=self.num_templates,
                num_prim_cells=self.num_prim_cells,
                init_temp=self.Tmax,
                final_temp=self.Tmin,
                num_temp=self.num_temps,
                num_steps_per_temp=self.num_steps,
                eci=self.eci)
            msg = 'Ground-state structures generated.'
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
        self.page.structure_generation_in_progress = False


class NewStructPage(Screen):
    _pop_up = None
    structure_generation_in_progress = False

    def on_enter(self):
        self.on_new_struct_type_update(self.ids.newStructTypeSpinner.text)

    def on_new_struct_type_update(self, text):
        inactive = get_color_from_hex(INACTIVE_TEXT_COLOR)
        active = get_color_from_hex(FOREGROUND_TEXT_COLOR)

        if text in ['Initial pool', 'Random structure']:
            self.ids.templateAtomsLabel.color = active
            self.ids.tempMaxLabel.color = inactive
            self.ids.tempMinLabel.color = inactive
            self.ids.numTempLabel.color = inactive
            self.ids.numSweepsLabel.color = inactive
            self.ids.eciFileLabel.color = inactive
            self.ids.numTemplateLabel.color = inactive
            self.ids.numPrimCells.color = inactive

            self.ids.loadTemplateAtoms.disabled = False
            self.ids.templateAtomsInput.disabled = False
            self.ids.tempMaxInput.disabled = True
            self.ids.tempMinInput.disabled = True
            self.ids.numTempInput.disabled = True
            self.ids.numSweepsInput.disabled = True
            self.ids.eciFileInput.disabled = True
            self.ids.randomizeCompositionSpinner.disabled = True
            self.ids.loadECIFile.disabled = True
            self.ids.numTemplateInput.disabled = True
            self.ids.numPrimCellsInput.disabled = True

        elif text == 'Probe structure':
            self.ids.tempMaxLabel.color = active
            self.ids.tempMinLabel.color = active
            self.ids.numTempLabel.color = active
            self.ids.numSweepsLabel.color = active
            self.ids.templateAtomsLabel.color = active
            self.ids.eciFileLabel.color = inactive
            self.ids.numTemplateLabel.color = inactive
            self.ids.numPrimCells.color = inactive

            self.ids.tempMaxInput.disabled = False
            self.ids.tempMinInput.disabled = False
            self.ids.numTempInput.disabled = False
            self.ids.numSweepsInput.disabled = False
            self.ids.loadTemplateAtoms.disabled = False
            self.ids.templateAtomsInput.disabled = False
            self.ids.eciFileInput.disabled = True
            self.ids.randomizeCompositionSpinner.disabled = True
            self.ids.loadECIFile.disabled = True
            self.ids.numTemplateInput.disabled = True
            self.ids.numPrimCellsInput.disabled = True

        elif text == 'Ground-state structure (fixed template)':
            self.ids.tempMaxLabel.color = active
            self.ids.tempMinLabel.color = active
            self.ids.numTempLabel.color = active
            self.ids.numSweepsLabel.color = active
            self.ids.eciFileLabel.color = active
            self.ids.templateAtomsLabel.color = active
            self.ids.numTemplateLabel.color = inactive
            self.ids.numPrimCells.color = inactive

            self.ids.tempMaxInput.disabled = False
            self.ids.tempMinInput.disabled = False
            self.ids.numTempInput.disabled = False
            self.ids.numSweepsInput.disabled = False
            self.ids.eciFileInput.disabled = False
            self.ids.randomizeCompositionSpinner.disabled = False
            self.ids.loadECIFile.disabled = False
            self.ids.loadTemplateAtoms.disabled = False
            self.ids.templateAtomsInput.disabled = False
            self.ids.numTemplateInput.disabled = True
            self.ids.numPrimCellsInput.disabled = True

        elif text == 'Ground-state structure (variable template)':
            self.ids.tempMaxLabel.color = active
            self.ids.tempMinLabel.color = active
            self.ids.numTempLabel.color = active
            self.ids.numSweepsLabel.color = active
            self.ids.eciFileLabel.color = active
            self.ids.numTemplateLabel.color = active
            self.ids.numPrimCells.color = active
            self.ids.templateAtomsLabel.color = inactive

            self.ids.tempMaxInput.disabled = False
            self.ids.tempMinInput.disabled = False
            self.ids.numTempInput.disabled = False
            self.ids.numSweepsInput.disabled = False
            self.ids.eciFileInput.disabled = False
            self.ids.randomizeCompositionSpinner.disabled = True
            self.ids.loadECIFile.disabled = False
            self.ids.numTemplateInput.disabled = False
            self.ids.numPrimCellsInput.disabled = False
            self.ids.loadTemplateAtoms.disabled = True
            self.ids.templateAtomsInput.disabled = True

    def dismiss_popup(self):
        self._pop_up.dismiss()
        self._pop_up = None

    def show_load_eci_dialog(self):
        content = LoadDialog(load=lambda path, filename: self.load(path, filename, 'eci'),
                             cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load ECI filename",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_load_template_dialog(self):
        content = LoadDialog(load=lambda path, filename: self.load(path, filename, 'template'),
                             cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load template atoms",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def load(self, path, filename, field='eci'):
        if field == 'eci':
            self.ids.eciFileInput.text = filename[0]
        elif field == 'template':
            self.ids.templateAtomsInput.text = filename[0]
        self.dismiss_popup()

    def load_eci(self, fname):
        with open(fname, 'r') as infile:
            eci = json.load(infile)
        return eci

    def get_eci_file(self):
        eci_file = self.ids.eciFileInput.text

        if eci_file == '':
            raise ValueError('No ECI file given')
        return eci_file

    def generate(self):
        from clease import NewStructures
        from ase.io import read

        if self.structure_generation_in_progress:
            # Don't allow user to initialize many threads
            # by successively clicking on the generate button
            return

        self.structure_generation_in_progress = True
        settings = App.get_running_app().root.settings

        if settings is None:
            msg = "Settings is not set. "
            msg += "Make sure Apply settings button was clicked.'"
            App.get_running_app().root.ids.status.text = msg
            self.structure_generation_in_progress = False
            return

        if self.ids.genNumberInput.text == '':
            generation_number = None
        else:
            generation_number = int(self.ids.genNumberInput.text)
        struct_per_gen = int(self.ids.structPerGenInput.text)

        try:
            generator = NewStructures(settings,
                                      generation_number=generation_number,
                                      struct_per_gen=struct_per_gen)

            fname = self.ids.templateAtomsInput.text
            if fname == '':
                msg = 'No Atoms template given. Using active template.'
                App.get_running_app().root.ids.status.text = msg
                atoms = settings.atoms.copy()
            else:
                if not os.path.exists(fname):
                    msg = f"Cannot find file {fname}"
                    App.get_running_app().root.ids.status.text = msg
                    self.structure_generation_in_progress = False
                    return

                atoms = read(fname)

            struct_type = self.ids.newStructTypeSpinner.text

            Tmin = float(self.ids.tempMinInput.text)
            Tmax = float(self.ids.tempMaxInput.text)
            num_temps = int(self.ids.numTempInput.text)
            num_steps = int(self.ids.numSweepsInput.text) * len(atoms)
            num_templates = int(self.ids.numTemplateInput.text)
            num_prim_cells = int(self.ids.numPrimCellsInput.text)

            if struct_type == 'Initial pool':
                msg = "Generating initial pool of structures..."
                App.get_running_app().root.ids.status.text = msg
                init_generator = InitialPoolGenerator()
                init_generator.generator = generator
                init_generator.atoms = atoms
                init_generator.status = App.get_running_app().root.ids.status
                init_generator.page = self
                Thread(target=init_generator.generate).start()
            elif struct_type == 'Random structure':
                msg = "Generating random structures..."
                App.get_running_app().root.ids.status.text = msg
                rnd_generator = RandomStructureGenerator()
                rnd_generator.generator = generator
                rnd_generator.atoms = atoms
                rnd_generator.status = App.get_running_app().root.ids.status
                rnd_generator.page = self
                Thread(target=rnd_generator.generate).start()
            elif struct_type == 'Probe structure':
                msg = 'Generating probe structures...'
                App.get_running_app().root.ids.status.text = msg
                prb_generator = ProbeStructureGenerator()
                prb_generator.atoms = atoms
                prb_generator.generator = generator
                prb_generator.status = App.get_running_app().root.ids.status
                prb_generator.Tmax = Tmax
                prb_generator.Tmin = Tmin
                prb_generator.num_temp = num_temps
                prb_generator.num_steps = num_steps
                prb_generator.page = self
                Thread(target=prb_generator.generate).start()
            elif struct_type == 'Ground-state structure (fixed template)':
                eci_file = self.get_eci_file()
                eci = self.load_eci(eci_file)
                msg = 'Generating ground-state structures...'
                App.get_running_app().root.ids.status.text = msg
                random_comp = self.ids.randomizeCompositionSpinner.text
                randomize = random_comp == 'Random composition'

                gs_generator = GSStructGenerator()
                gs_generator.atoms = atoms
                gs_generator.generator = generator
                gs_generator.status = App.get_running_app().root.ids.status
                gs_generator.Tmax = Tmax
                gs_generator.Tmin = Tmin
                gs_generator.num_temps = num_temps
                gs_generator.num_steps = num_steps
                gs_generator.eci = eci
                gs_generator.randomize = randomize
                gs_generator.page = self

                Thread(target=gs_generator.generate).start()
            elif struct_type == 'Ground-state structure (variable template)':
                eci_file = self.get_eci_file()
                eci = self.load_eci(eci_file)
                msg = 'Generating ground-state structures...'
                App.get_running_app().root.ids.status.text = msg

                gs_generator = ExaustiveGSStructGenerator()
                gs_generator.generator = generator
                gs_generator.status = App.get_running_app().root.ids.status
                gs_generator.Tmax = Tmax
                gs_generator.Tmin = Tmin
                gs_generator.num_temps = num_temps
                gs_generator.num_steps = num_steps
                gs_generator.eci = eci
                gs_generator.num_templates = num_templates
                gs_generator.num_prim_cells = num_prim_cells
                gs_generator.page = self

                Thread(target=gs_generator.generate).start()
        except (RuntimeError, ValueError) as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
            self.structure_generation_in_progress = False

    def import_structures(self):
        from ase.io import read
        from clease import NewStructures
        init = self.ids.initStructInput.text

        if init == '' or not os.path.exists(init):
            msg = f"Cannot find initial structure {init}"
            App.get_running_app().root.ids.status.text = msg
            return

        try:
            init_struct = read(init)

            final = self.ids.finalStructInput.text

            if final == '':
                final_struct = None
            else:
                if not os.path.exists(final):
                    msg = f"Cannot find final structure {final}"
                    App.get_running_app().root.ids.status.text = msg
                    return
                final_struct = read(final)

            settings = App.get_running_app().root.settings

            if settings is None:
                msg = 'Settings is not set. '
                msg += 'Make sure Apply settings button was clicked.'
                App.get_running_app().root.ids.status.text = msg
                return

            if self.ids.genNumberInput.text == '':
                generation_number = None
            else:
                generation_number = int(self.ids.genNumberInput.text)
            struct_per_gen = int(self.ids.structPerGenInput.text)

            generator = NewStructures(settings,
                                      generation_number=generation_number,
                                      struct_per_gen=struct_per_gen)

            status = App.get_running_app().root.ids.status
            status.text = 'Inserting structures...'

            # The argument passed is a trajectory file
            if final.endswith('.traj') and init.endswith('.traj'):
                kwargs = {'traj_init': init, 'traj_final': final, 'cb': InsertStructureCB(status)}
                Thread(target=generator.insert_structures, kwargs=kwargs).start()
            elif init.endswith('.traj'):
                kwargs = {'traj_init': init, 'cb': InsertStructureCB(status)}
                Thread(target=generator.insert_structures, kwargs=kwargs).start()
            else:
                generator.insert_structure(init_struct=init_struct, final_struct=final_struct)
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)

    def load_structures(self, path, filename, is_init):
        if len(filename) == 0:
            App.get_running_app().root.ids.status.text = 'No file selecton...'
            self.dismiss_popup()
            return

        if is_init:
            self.ids.initStructInput.text = filename[0]
        else:
            self.ids.finalStructInput.text = filename[0]
        self.dismiss_popup()

    def show_load_init_struct_dialog(self):
        content = LoadDialog(load=lambda path, filename: self.load_structures(path, filename, True),
                             cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load initial structure",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_load_final_struct_dialog(self):
        content = LoadDialog(
            load=lambda path, filename: self.load_structures(path, filename, False),
            cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load final structure",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def todict(self):
        return {
            'gen_number': self.ids.genNumberInput.text,
            'struct_per_gen': self.ids.structPerGenInput.text,
            'init_struct': self.ids.initStructInput.text,
            'final_struct': self.ids.finalStructInput.text,
            'min_temp': self.ids.tempMinInput.text,
            'max_temp': self.ids.tempMaxInput.text,
            'num_temps': self.ids.numTempInput.text,
            'num_sweeps': self.ids.numSweepsInput.text,
            'eci_file': self.ids.eciFileInput.text,
            'template_file': self.ids.templateAtomsInput.text,
            'generation_scheme': self.ids.newStructTypeSpinner.text,
            'num_templates': self.ids.numTemplateInput.text
        }

    def from_dict(self, data):
        self.ids.genNumberInput.text = data.get('gen_number', '')
        self.ids.structPerGenInput.text = data.get('struct_per_gen', '5')
        self.ids.initStructInput.text = data.get('init_struct', '')
        self.ids.finalStructInput.text = data.get('final_struct', '')
        self.ids.tempMinInput.text = data.get('min_temp', '1')
        self.ids.tempMaxInput.text = data.get('max_temp', '10000')
        self.ids.numTempInput.text = data.get('num_temps', '100')
        self.ids.numSweepsInput.text = data.get('num_sweeps', '100')
        self.ids.eciFileInput.text = data.get('eci_file', '')
        self.ids.templateAtomsInput.text = data.get('template_file', '')
        self.ids.newStructTypeSpinner.text = data.get('generation_scheme', 'Random structure')
        self.ids.numTemplateInput.text = data.get('num_templates', '1')



from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty
import os
import json
from ase.data import chemical_symbols
from clease import __version__ as version
from clease.gui import backup_folder


class SGCEditor(FloatLayout):
    close = ObjectProperty(None)
    data = {}
    backup_file = f"sgcEditorBck{version}.json"

    def __init__(self, **kwargs):
        symbols = kwargs.pop('symbols')
        chem_pot = kwargs.pop('chem_pot')
        FloatLayout.__init__(self, **kwargs)

        self.ids.symbolInput.text = symbols
        self.ids.chemPotInput.text = chem_pot

        if not backup_folder.exists():
            backup_folder.mkdir()
        self.load_values()

    def backup(self):
        self.data['symbols'] = self.ids.symbolInput.text
        self.data['chem_pot'] = self.ids.chemPotInput.text
        with open(self.fname, 'w') as f:
            json.dump(self.data, f)

    def load_values(self):
        if os.path.exists(self.fname):
            with open(self.fname, 'r') as f:
                self.data = json.load(f)

            self.ids.symbolInput.text = self.data['symbols']
            self.ids.chemPotInput.text = self.data['chem_pot']

    @property
    def fname(self):
        return str(backup_folder / self.backup_file)

    def symbols_are_valid(self, symbs):
        for s in symbs:
            if s not in chemical_symbols:
                return False
        return True

    def parse_symbols(self):
        symbs = self.ids.symbolInput.text
        remove = ['[', ']', '(', ')']
        for r in remove:
            symbs = symbs.replace(r, '')
        symbs = [x.strip() for x in symbs.split(',')]
        return symbs

    def parse_chem_pot(self):
        chem_pot = self.ids.chemPotInput.text
        chem_pot = chem_pot.replace('{', '')
        chem_pot = chem_pot.replace('}', '')

        chem_pot_dict = {}
        for item in chem_pot.split(','):
            splitted = item.split(':')
            if len(splitted) != 2:
                continue
            key = splitted[0].strip()
            value = float(splitted[1])
            chem_pot_dict[key] = value
        return chem_pot_dict



from ase.atoms import Atoms, Cell
import spglib
import numpy as np
from scipy.optimize import linear_sum_assignment
from typing import Tuple
from ase.geometry import find_mic

__all__ = ('TransformInfo', 'StructureMapper')


class TransformInfo(object):
    """
    Class for holding information about snap transformation
    """

    def __init__(self):
        self.displacements = []
        self.dispVec = []
        self.strain = []

    def todict(self) -> dict:
        return {'displacements': list(self.displacements), 'strain': list(self.strain)}


class StructureMapper(object):
    """
    Class that refines a relaxed structure to a corresponding "ideal"
    lattice.

    :param symprec: Symmetry precision (see refine_cell in spglib)
    :param angle_tol: Angle tolerance (see refine_cell in spglib)
    """

    def __init__(self, symprec: float = 0.1, angle_tol: float = 5.0):
        self.symprec = symprec
        self.angle_tol = angle_tol

    def refine(self, atoms: Atoms) -> Atoms:
        """
        Refine a relaxed atoms object. This method relies on symmetry finding.
        It handles distortions to some extent, but for heavily distorted
        systems, it will most likely fail.

        :param atoms: Relaxed atoms object
        """
        cell = np.array(atoms.get_cell())
        scaled_pos = atoms.get_scaled_positions()
        numbers = np.array(atoms.numbers)

        result = spglib.refine_cell((cell, scaled_pos, numbers),
                                    symprec=self.symprec,
                                    angle_tolerance=self.angle_tol)

        if result is None:
            raise RuntimeError("Could not refine cell")

        lattice, scaled_pos, numbers = result
        return Atoms(numbers=numbers, cell=lattice, scaled_positions=scaled_pos, pbc=[1, 1, 1])

    def strain(self, cell1: Cell, cell2: Cell) -> np.ndarray:
        """
        Calculate the strain tensor to map cell1 into cell2.
        In the following the 3x3 matrix where each cell vector is a
        column, is denoted C_1 and C_2 for cell1 and cell2, respectively.
        There is a transformation matrix P such that C_2 = PC_1. Thus,
        P = C_2C_1^{-1}. The strain tensor is defined as E = 0.5*(P^TP - I),
        where I is the identity matrix.

        :param cell1: First cell
        :param cell2: Second cell
        """
        P = cell2.dot(np.linalg.inv(cell1))
        return 0.5 * (P.T.dot(P) - np.eye(3))

    def snap_to_lattice(self, atoms: Atoms, template: Atoms) -> Tuple[Atoms, TransformInfo]:
        """
        Snaps atoms to an ideal lattice. If the number of atoms in the
        relaxed atoms object is less than the number of atoms in the template,
        the method will add vacancies on sites that are not matched.

        :param atoms: Relaxed atoms object. Note that the cell and atomic
            positions will be altered.
        :param template: Ideal lattice. On return, the symbols corresponds
            to the mapped symbols in atoms
        """
        strain = self.strain(atoms.get_cell(), template.get_cell())
        atoms.set_cell(template.get_cell(), scale_atoms=True)

        # Build distance matrix using MIC
        distances = np.zeros((len(atoms), len(template)))
        cell = template.get_cell()
        all_dist_vec = []
        for i in range(len(atoms)):
            for j in range(len(template)):
                d = template[j].position - atoms[i].position
                all_dist_vec.append(d)
        distVec, micDist = find_mic(all_dist_vec, cell)
        distances = micDist.reshape((len(atoms), len(template)))

        # Snap atoms to ideal sites such that collective distance
        # is minimized
        row, assignment = linear_sum_assignment(distances)
        displacements = distances[row, assignment]
        template.numbers = np.zeros(len(template))
        template.numbers[assignment] = atoms.numbers

        transform_info = TransformInfo()
        transform_info.displacements = displacements
        transform_info.strain = strain
        transform_info.dispVec = np.array([
            distVec[i * len(template) + j] for i in range(len(atoms)) for j in range(len(template))
        ])
        return template, transform_info



from .mc_constraint import MCConstraint
from .constrain_swap_by_basis import ConstrainSwapByBasis
from .constrain_element_insert_by_basis import ConstrainElementInserts
from .fixed_element import FixedElement
from .collective_variable_constraint import CollectiveVariableConstraint
from .pair_constraint import PairConstraint
from .fixed_indices import FixedIndices

__all__ = [
    "MCConstraint",
    "ConstrainSwapByBasis",
    "ConstrainElementInserts",
    "FixedElement",
    "CollectiveVariableConstraint",
    "PairConstraint",
    "FixedIndices",
]



from itertools import product, chain
from copy import deepcopy
import numpy as np
from ase.geometry import wrap_positions

from .cluster_fingerprint import ClusterFingerprint
from .cluster import Cluster
from .cluster_list import ClusterList
from .cluster_generator import ClusterGenerator
from .utils import name_clusters, size

__all__ = ('ClusterManager',)


class ClusterManager:
    """
    Manager for construction of all cluster.

    Parameters:

    prim_cell: ase.Atoms
        Primitive cell
    """

    def __init__(self, prim_cell):
        self.generator = ClusterGenerator(prim_cell)
        self.clusters = ClusterList()

    def __eq__(self, other):
        return self.clusters == other.clusters and \
            self.generator == other.generator

    def build(self, max_size=4, max_cluster_dia=4.0):
        """
        Construct all clusters.

        Parameters:

        max_size: int
            Maximum number of atoms in a cluster

        max_cluster_dia: float
            Maximum distance between two atoms in a cluster
        """
        if isinstance(max_cluster_dia, float):
            max_cluster_dia = [max_cluster_dia for _ in range(max_size + 1)]
        num_lattices = range(len(self.generator.prim))
        cluster_size = range(2, max_size + 1)
        all_fps = []
        names = []
        all_clusters = []
        all_eq_sites = []
        lattices = []
        diameters = []
        sizes = []
        for latt, s in product(num_lattices, cluster_size):
            clusters, fps = self.generator.generate(s, max_cluster_dia[s], ref_lattice=latt)

            eq_sites = []
            for c in clusters:
                eq_sites.append(self.generator.equivalent_sites(c[0]))

            all_fps += fps
            all_clusters += clusters
            all_eq_sites += eq_sites
            lattices += [latt] * len(clusters)
            sizes += [s] * len(clusters)
            diameters += [2 * np.sqrt(fp[0]) for fp in fps]

        names = self._get_names(all_fps)
        # Transfer to the cluster list
        zipped = zip(names, sizes, diameters, all_fps, all_clusters, all_eq_sites, lattices)
        for n, s, d, fp, c, eq, l in zipped:
            cluster = Cluster(name=n,
                              size=s,
                              diameter=d,
                              fingerprint=fp,
                              ref_indx=-1,
                              indices=c,
                              equiv_sites=eq,
                              trans_symm_group=l)
            self.clusters.append(cluster)

        # Add singlets and empty
        for i in range(len(self.generator.prim)):
            self.clusters.append(
                Cluster(name='c1',
                        size=1,
                        diameter=0.0,
                        fingerprint=ClusterFingerprint([1.0]),
                        ref_indx=-1,
                        indices=[[[0, 0, 0, i]]],
                        equiv_sites=[],
                        trans_symm_group=i))
            self.clusters.append(
                Cluster('c0',
                        size=0,
                        diameter=0.0,
                        fingerprint=ClusterFingerprint([0.0]),
                        ref_indx=-1,
                        indices=[],
                        equiv_sites=[],
                        trans_symm_group=i))

    def _get_names(self, all_fps):
        """
        Give a consistent name to all clusters

        Parameter:

        all_fps: list of ClusterFingerPrint
            A list with all the cluster fingerprints
        """
        # The following one-liner will do the job when we
        # decide to update cluster names from the legacy ones
        # where names where per by size
        # names = name_clusters(all_fps)

        sizes = set([size(fp) for fp in all_fps])
        names = [None for _ in all_fps]
        for s in sizes:
            fps = []
            indices = []
            for i, fp in enumerate(all_fps):
                if size(fp) == s:
                    indices.append(i)
                    fps.append(fp)
            names_per_size = name_clusters(fps)
            for i, n in zip(indices, names_per_size):
                names[i] = n
        return names

    def _ref_indices(self, kdtree):
        """
        Return all reference indices

        Parameters:

        kdtree: KDTree or cKDTree
            A KDtree representation of all atomic positions
        """
        ref_indices = []
        for i in range(len(self.generator.prim)):
            d, i = kdtree.query(self.generator.cartesian([0, 0, 0, i]))
            ref_indices.append(i)
        return ref_indices

    def info_for_template(self, template):
        """
        Specialise the cluster information to a template

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        unique = self.unique_four_vectors()
        lut = self.fourvec_to_indx(template, unique)
        ref_indices = [lut[(0, 0, 0, i)] for i in range(self.generator.num_sub_lattices)]

        cluster_int = deepcopy(self.clusters)
        for cluster in cluster_int:
            if cluster.size == 0:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            elif cluster.size == 1:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            else:
                cluster.indices = self.generator.to_atom_index(cluster.indices, lut)
                cluster.ref_indx = int(ref_indices[cluster.group])
        return cluster_int

    def unique_four_vectors(self):
        """
        Return a list with all unique 4-vectors consituting
        the clusters
        """
        unique = set()
        for c in self.clusters:
            if c.size == 0:
                continue
            flat = list(chain(*c.indices))
            indices = list(map(tuple, flat))
            unique.update(indices)
        return unique

    def get_figures(self):
        """
        Return a list of atoms object representing the clusters
        """
        return self.clusters.get_figures(self.generator)

    def create_four_vector_lut(self, template):
        """
        Construct a lookup table (LUT) for the index in template given the
        wrapped vector

        Parameter:

        template: Atoms
            Atoms object to use when creating the lookup table (LUT)
        """
        lut = {}
        pos = template.get_positions().copy()
        for i in range(pos.shape[0]):
            if template[i].tag >= self.generator.num_sub_lattices:
                continue
            vec = self.generator.get_four_vector(pos[i, :], template[i].tag)
            lut[tuple(vec)] = i
        return lut

    def fourvec_to_indx(self, template, unique):
        cell = template.get_cell()
        pos = np.zeros((len(unique), 3))
        for i, u in enumerate(unique):
            cart_u = self.generator.cartesian(u)
            pos[i, :] = cart_u

        pos = wrap_positions(pos, cell)
        unique_indices = []
        for i in range(pos.shape[0]):
            diff_sq = np.sum((pos[i, :] - template.get_positions())**2, axis=1)
            unique_indices.append(np.argmin(diff_sq))
        return dict(zip(unique, unique_indices))

    def translation_matrix(self, template):
        """
        Construct the translation matrix

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        trans_mat = []
        unique = self.unique_four_vectors()
        cell = template.get_cell()

        unique_indx = self.fourvec_to_indx(template, unique)

        lut = self.create_four_vector_lut(template)
        indices = [0 for _ in range(len(unique))]
        cartesian = np.zeros((len(unique), 3))

        # Make a copy of the positions to avoid atoms being translated
        tmp_pos = template.get_positions().copy()
        unique_npy = np.array(list(unique))
        translated_unique = np.zeros_like(unique_npy)
        for atom in template:
            if atom.tag >= self.generator.num_sub_lattices:
                trans_mat.append({})
                continue

            vec = self.generator.get_four_vector(tmp_pos[atom.index, :], atom.tag)

            translated_unique[:, :3] = unique_npy[:, :3] + vec[:3]
            translated_unique[:, 3] = unique_npy[:, 3]
            cartesian = self.generator.cartesian(translated_unique)
            cartesian = wrap_positions(cartesian, cell)

            four_vecs = self.generator.get_four_vector(cartesian, translated_unique[:, 3])
            N = four_vecs.shape[0]
            indices = [lut[tuple(four_vecs[i, :])] for i in range(N)]

            trans_mat.append(dict(zip([int(unique_indx[u]) for u in unique], indices)))
        return trans_mat



from copy import deepcopy
import numpy as np
from scipy.spatial import cKDTree as KDTree
from ase.geometry import wrap_positions
from .cluster_list import ClusterList

__all__ = ('AtomsNotContainedInLargeCellError', 'ClusterInfoMapper')


class AtomsNotContainedInLargeCellError(Exception):
    pass


class ClusterInfoMapper:
    """
    Class for accelerating the construction of cluster descriptors when
    the cluster descriptors are already known for a larger cell

    :param atoms: ASE Atoms object
        Large atoms cell

    :param trans_matrix: list of dict
        Translation matrix for the large cell

    :param cluster_list: ClusterList
        Cluster info the for the large cell
    """

    def __init__(self, atoms, trans_matrix, cluster_list):
        self.atoms = atoms
        self.trans_matrix = trans_matrix
        self.cluster_list = cluster_list
        self.tree = KDTree(self.atoms.get_positions())

    def _map_indices(self, small_atoms):
        """Map indices from large atoms to small."""
        tree = KDTree(small_atoms.get_positions())

        # Positions in large cell
        large_pos = self.atoms.get_positions()

        # Wrap these positions into the small cell
        large_pos = wrap_positions(large_pos, small_atoms.get_cell())

        # loop through small Atoms and find the closest atoms in large cell
        dist, index_map = tree.query(large_pos)
        assert np.all(dist < 1E-6)

        if len(list(set(index_map))) != len(small_atoms):
            raise AtomsNotContainedInLargeCellError("All indices not covered")
        return index_map

    def _map_cluster_info(self, index_map):
        """
        Map cluster info of the large cell to the corresponding cluster info
        of the small cell
        """
        new_cluster_list = ClusterList()

        for cluster in self.cluster_list:
            new_cluster = deepcopy(cluster)
            new_cluster.ref_indx = int(index_map[cluster.ref_indx])
            new_cluster.indices = [[int(index_map[x]) for x in sub] for sub in cluster.indices]
            new_cluster_list.append(new_cluster)
        return new_cluster_list

    def _map_trans_matrix(self, index_map):
        """Map translation matrix."""
        unique = list(set(index_map))
        new_tm = []

        for row in unique:
            row_large = np.where(index_map == row)[0][0]
            new_tm.append({
                int(index_map[k]): int(index_map[v])
                for k, v in self.trans_matrix[row_large].items()
            })
        return new_tm

    def map_info(self, small_atoms):
        """Map cluster info and translation matrix from the large host."""
        index_map = self._map_indices(small_atoms)
        new_info = self._map_cluster_info(index_map)
        new_tm = self._map_trans_matrix(index_map)
        return new_info, new_tm



import logging
import numpy as np
from scipy.optimize import minimize

from .regression import LinearRegression

logger = logging.getLogger(__name__)

__all__ = ('EigenDecomposition', 'GeneralizedRidgeRegression')


class EigenDecomposition:

    def __init__(self) -> None:
        self.eigvals = None
        self.eigvec = None

    def decompose(self, X: np.ndarray) -> None:
        """
        Performs eigen decomposition of a symmetric matrix X
        """
        self.eigvals, self.eigvec = np.linalg.eigh(X)


class GeneralizedRidgeRegression(LinearRegression):
    """
    GeneralizedRidgeRegression performs a ridge regression where each feature
    has its own penalization term. The optimimal value for each penalization
    coefficient is optimed by minimizing the generalized CV score. The method
    is based on the ideas presented in

    Golub, G.H., Heath, M. and Wahba, G., 1979.
    Generalized cross-validation as a method for choosing a good ridge
    parameter. Technometrics, 21(2), pp.215-223.

    :param alpha: Initial guess for the penalization parameters. If not
        given, the initial guess will be set to zero for all coefficients.
        Internally, a local optimization method is used to find the optimal
        penalization values. Thus, in case there are multiple local minima
        (which it normally is), the result from fit may vary when different
        initial guesses for alpha is provided.
    """

    def __init__(self, alpha: np.ndarray = None) -> None:
        super().__init__()
        self.alpha = alpha

        # Dictionary that will be populated with results from the
        # hyperparameter optimization
        self.opt_result = {}

        # Dictionary that is unpacked and passed to scipy.optimize.minimize
        # (e.g. minimize(**self.minimize_args)). See doc of
        # scipy.optimize.minimize for a detailed description of possible
        # options
        self.minimize_args = {}

    @staticmethod
    def _A(X: np.ndarray, eigen: EigenDecomposition, alpha: np.ndarray) -> np.ndarray:
        """
        Helper method to form the matrix
        X(X^TX + N*diag(alpha))^{-1}X^T, where N
        is the number of rows in X

        :param X: Design matrix
        :param eigen: Eigen decomposition of X^TX
        :param alpha: Vector with penalization values
        """
        N = X.shape[0]
        return np.linalg.multi_dot(
            [X, eigen.eigvec,
             np.diag(1.0 / (eigen.eigvals + N * alpha)), eigen.eigvec.T, X.T])

    def _coeff(self, X: np.ndarray, y: np.ndarray, eigen: EigenDecomposition) -> np.ndarray:
        """
        Calculate the coefficients using self.alpha as regularization

        :param X: Design matrix
        :param y: Target values
        :param eigen: Eigen decomposition of X^TX
        """
        N = X.shape[0]
        return np.linalg.multi_dot(
            [eigen.eigvec,
             np.diag(1.0 / (eigen.eigvals + N * self.alpha)), eigen.eigvec.T, X.T, y])

    @staticmethod
    def _grad_A(X: np.ndarray, eigen: EigenDecomposition, alpha: np.ndarray,
                component: int) -> np.ndarray:
        """
        Return a matrix representing the gradient of the A matrix.
        See the `_A` method.

        :param X: Design matrix
        :param eigen: Eigen decomposition of X^TX
        :param alpha: Vector with penalization value
        :param component: Integer specifying which component of
            the gradient that should be calculated
        """
        N = X.shape[0]
        XdotP = np.dot(X, eigen.eigvec[:, component])
        weight = N / (eigen.eigvals[component] + N * alpha[component])**2
        return -np.outer(XdotP, XdotP.T) * weight

    def _gcv_squared(self, alpha: np.ndarray, X: np.ndarray, y: np.ndarray,
                     eigen: EigenDecomposition) -> float:
        """
        Return the generalized CV score of a fit

        :param alpha: Vector of length M with regularization parameters
        :param X: Design matrix (N x M)
        :param y: Target values of length N
        :param eigen: EigenDecomposition of X^TX
        """
        A = self._A(X, eigen, alpha)
        N = len(y)
        mse = np.mean((y - A.dot(y))**2)
        denum = (1.0 - np.trace(A) / N)**2
        return mse / denum

    def _grad_gcv_squared(self, alpha: np.ndarray, X: np.ndarray, y: np.ndarray,
                          eigen: EigenDecomposition) -> np.ndarray:
        """
        Return the gradient of the squared generalized CV score.

        :param alpha: Vector of length M with regularization parameters
        :param X: Design matrix (N x M)
        :param y: Target values of length N
        :param eigen: EigenDecomposition of X^TX
        """
        grad = np.zeros(len(alpha))
        A = self._A(X, eigen, alpha)
        dev = y - A.dot(y)
        mse = np.mean(dev**2)
        N = len(y)
        denum = (1.0 - np.trace(A) / N)
        for i in range(len(grad)):
            gradA = self._grad_A(X, eigen, alpha, i)
            new_grad = -2.0 * np.sum(np.outer(dev, y) * gradA) / (N * denum**2)

            # There is an additional contribution for the diagonal terms
            # in gradA
            new_grad += 2.0 * mse * np.trace(gradA) / (N * denum**3)
            grad[i] = new_grad
        return grad

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Fit X*coeff = y. The hyper parameters are optimized internally.

        :param X: Design matrix
        :param y: Target values
        """
        if self.alpha is None:
            self.alpha = np.zeros(X.shape[1])
        eigen = EigenDecomposition()
        eigen.decompose(X.T.dot(X))

        def cost_func(alpha):
            # Make sure that we always have positive penalization
            x = np.exp(alpha)
            return self._gcv_squared(x, X, y, eigen)

        def jac_cost_func(alpha):
            x = np.exp(alpha)
            grad = self._grad_gcv_squared(x, X, y, eigen)
            return grad * x

        res = minimize(cost_func, np.log(self.alpha), jac=jac_cost_func, **self.minimize_args)
        self.alpha = np.exp(res.x)

        if not res.success:
            logger.warning("Failed to find optimal regularization parameters.")
            logger.warning(res.message)

        # Sanity checks
        A = self._A(X, eigen, self.alpha)
        eff_num_params = np.trace(A)
        if eff_num_params < 0.0:
            logger.warning(("Warning! The effective number of parameters is negative. "
                            "Try to change the initial guess for alpha."))
        logger.info('Best GCV: %.3f', np.sqrt(res.fun))
        coeff = self._coeff(X, y, eigen)
        self.opt_result = {
            'gcv': np.sqrt(res.fun),
            'scipy_opt_res': res,
            'gcv_dev': (y - X.dot(coeff)) / (1.0 - eff_num_params / X.shape[0]),
            'press_dev': (y - X.dot(coeff)) / (1.0 - np.diag(A))
        }
        return coeff



# pylint: disable=undefined-variable
from .regression import *
from .generalized_ridge_regression import *
from .bayesian_compressive_sensing import *
from .ga_fit import *
from .constrained_ridge import *
from .physical_ridge import *
from .sequential_cluster_ridge import *

__all__ = (regression.__all__ + generalized_ridge_regression.__all__ +
           bayesian_compressive_sensing.__all__ + ga_fit.__all__ + physical_ridge.__all__ +
           sequential_cluster_ridge.__all__ + constrained_ridge.__all__)



import numpy as np
from ase.units import kB
import time


class MetaDynGuiUpdater(object):

    def __init__(self, meta_dyn_page=None, meta_dyn_sampler=None):
        self.page = meta_dyn_page
        self.sampler = meta_dyn_sampler

    def __call__(self):
        xmin = self.sampler.bias.xmin
        xmax = self.sampler.bias.xmax
        x = np.linspace(xmin, xmax, 100)
        beta = 1.0 / (kB * self.sampler.mc.T)
        betaG = [-self.sampler.bias.evaluate(y) * beta for y in x]
        self.page.update_free_energy_plot(x, betaG)

        visits = [self.sampler.visit_hist.evaluate(y) for y in x]
        self.page.update_visit_plot(x, visits, self.sampler.flat_limit)
        time.sleep(0.01)
